# nodus - Job Management Framework
# Author: Manuel Blanco Valentin
# Email: manuel.blanco.valentin@gmail.com
# Created: 2024-12-25
#
# Description:
# This file contains the some utilities for the Nodus job management framework.
#
# File: utils.py
# Purpose:
# - Contains some basic functionalities used by the Nodus framework.

""" Basic imports """
import os
import re

""" Loggers """
import logging
from datetime import datetime

""" Subprocess to check if pd is running using ps (if not linux) """
import subprocess

""" Platform to determine the current system """
import platform

""" Numpy """
import numpy as np

""" Import nodus """
import nodus

""" Printers """
import logging
from datetime import datetime


""" Get nodus version """
def get_nodus_version():
    """Get the version of the Nodus framework."""
    return nodus.__version__


class CustomFormatter(logging.Formatter):
    def __init__(self, *args, level_formatter = "{level_name}", time_formatter = "{current_time:<8}", **kwargs):
        super().__init__(*args, **kwargs)
        self.last_level = None
        self.level_formatter = level_formatter
        self.time_formatter = time_formatter
        self.print_level = True

    def format(self, record):

        # Get the current time
        current_time = datetime.now().strftime('%H:%M:%S.%f')[:-4]

        # Format the log level and message
        level_name = record.levelname.replace('WARNING', 'WARN').replace('ERROR', 'ERR')
        message = record.getMessage()

        # Check if level changed 
        #print_level = self.last_level != level_name
        #self.last_level = level_name            

        # Format the log level and time 
        level_name = self.level_formatter.format(level_name = level_name)
        current_time = self.time_formatter.format(current_time = current_time)

        # Prepare the basic log entry (level, time, message)
        log_entry = f'╰ {level_name}┤ {current_time} │ - {message}'

        if self.print_level:
            return f'{log_entry}'
        else:
            # Otherwise, just print the time with a space for level
            #level_name = f"{'':<5}"  # Empty string for level
            level_name = self.level_formatter.format(level_name = "")
            space = ' ' * (len(level_name))  # Space for level
            return f' {space} │ {current_time} │ - {message}'

""" Custom logger """
class CustomLogger:
    def __init__(self, dir):
        self.last_level = None
        self.last_date = None
        self.formatter = None
        self.log_file_path = None

        # Get the maximum length allowed for the log message (which means, for the level name, and the date)
        self.__get_max_lengths()

        # Set up logger with both console and file logging
        self.logger = self.setup_logger(dir = dir, level_formatter = self.fmt_level, time_formatter = self.fmt_time, log_to_file=True)

    def __get_max_lengths(self):
        level_names = ['INFO', 'WARN', 'ERR', 'DBG', 'CRIT', 'LOG']
        self.max_level_length = max(max([len(level) for level in level_names]), 5)
        class fmt_level:
            def __init__(self, max_length):
                self.max_length = max_length
            def format(self, level_name = ""):
                return level_name + ' ' + (self.max_length-len(level_name))*"─"
        self.fmt_level = fmt_level(self.max_level_length)
        self.fmt_time = "{current_time:<8}"
    
    def setup_logger(self, dir = '.', log_to_file=False, log_file_path=None, **kwargs):
        # Set up the logger
        logger = logging.getLogger('nodus')
        logger.setLevel(logging.DEBUG)

        # Formatter for the log messages
        self.formatter = CustomFormatter('%(message)s', **kwargs)

        # Console handler to print to stdout
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(self.formatter)
        logger.addHandler(console_handler)

        # File handler to log to a file, if enabled
        if log_to_file:
            log_file_path = log_file_path or os.path.join(dir, f'nodus_{datetime.now().strftime("%Y%m%d_%H%M%S")}.log')
            self.log_file_path = log_file_path
            file_handler = logging.FileHandler(log_file_path)
            file_handler.setFormatter(self.formatter)
            print(f'Logging to file: {log_file_path}')
            logger.addHandler(file_handler)

        return logger

    # Function to get the current date and time
    def __get_datetime(self):
        # Get the current date and time
        current_date = datetime.now().strftime('%d/%b/%Y')
        current_time = datetime.now().strftime('%H:%M:%S.%f')[:-4]
        return current_date, current_time

    # Function to change the formatter of the logger handlers
    def __change_handlers_formatter(self, formatter):
        # Loop thru logger handlers
        for handler in self.logger.handlers:
            handler.setFormatter(formatter)

    def __check_date(self, current_date):
        # Handle date change
        print_cross = True
        if self.last_date != current_date:
            date_change = f'-- Date: {current_date}'
            self.last_date = current_date
            # Set None formatters just to print the date 
            self.__change_handlers_formatter(None)
            # Print the date
            self.logger.info(date_change)
            # Set the formatter back
            fmt = CustomFormatter('%(message)s', level_formatter = self.fmt_level, time_formatter = self.fmt_time)
            self.__change_handlers_formatter(fmt)
            # Assign the variable print_cross to False 
            print_cross = False
        return print_cross

    def __check_level(self, level_name, print_cross = True, print_header = False):
        # Check if level changed 
        old_level = str(self.last_level)
        print_top_border = (self.last_level != level_name) | print_header
        self.last_level = level_name

        # Define border characters
        bchar = ['┬','┼']
        bchar3 = ['╮','┤']

        # If we need to print the top border, do it here
        if print_top_border:
            # Draw a box around the log entry for the first message of the level
            border_top = '╭' + '─' * (self.max_level_length + 2) + bchar[int(print_cross)] + '─' * (8 + 2 + 3) + bchar3[int(print_cross)]
            # Set None formatters just to print the border
            self.__change_handlers_formatter(None)
            # Print the border
            self.logger.info(border_top)
            # Set the formatter back
            fmt = CustomFormatter('%(message)s', level_formatter = self.fmt_level, time_formatter = self.fmt_time)
            fmt.print_level = True
            self.__change_handlers_formatter(fmt)
        else:
            # Set the formatter back
            fmt = CustomFormatter('%(message)s', level_formatter = self.fmt_level, time_formatter = self.fmt_time)
            fmt.print_level = False
            self.__change_handlers_formatter(fmt)
        return print_top_border

    def _preprints(self, level, print_header = False):
        # Get the current date and time
        current_date, current_time = self.__get_datetime()
        print_cross = self.__check_date(current_date)

        # Check if the level changed. If it did, we need to print the top border
        print_top_border = self.__check_level(level, print_cross, print_header)

    def info(self, msg, print_header = False, **kwargs):
        self._preprints('INFO', print_header = print_header)
        self.logger.info(msg, **kwargs)

    def error(self, msg, print_header = False, **kwargs):
        self._preprints('ERR', print_header = print_header)
        self.logger.error(msg, **kwargs)

    def warn(self, msg, print_header = False, **kwargs):
        self._preprints('WARN', print_header = print_header)
        self.logger.warning(msg, **kwargs)
    
    def debug(self, msg, print_header = False, **kwargs):
        self._preprints('DBG', print_header = print_header)
        self.logger.debug(msg, **kwargs)
    
    def critical(self, msg, print_header = False, **kwargs):
        self._preprints('CRIT', print_header = print_header)
        self.logger.critical(msg, **kwargs)
    
    def custom(self, level, msg, print_header = False, **kwargs):
        # Ensure level is at max 4 characters
        level = level[:4]

        CUSTOM_LEVELV_NUM = 9 
        logging.addLevelName(CUSTOM_LEVELV_NUM, level)
        def _custom(self, message, *args, **kws):
            # Yes, logger takes its '*args' as 'args'.
            self._log(CUSTOM_LEVELV_NUM, message, args, **kws) 
        logging.Logger.custom = _custom

        self._preprints(level, print_header = print_header)
        self.logger.custom(msg, **kwargs)
    
    

    def close(self):
        level_name = self.fmt_level.format(level_name = "")
        space = ' ' * (len(level_name) + 2)  # Space for level
        # Print the last bottom border 
        border_bottom = space + '╰' + '─' * (8 + 5) + '╯'
        # Set formatters to None
        self.__change_handlers_formatter(None)
        # Print the bottom border
        self.logger.info(border_bottom)
        # Print a goodbye message with the current date and time
        current_date = datetime.now().strftime('%d/%b/%Y')
        current_time = datetime.now().strftime('%H:%M:%S.%f')[:-4]
        self.logger.info(f'-- Nodus closing @ [{current_date} {current_time}] 👋')
        # Set the formatter back
        fmt = CustomFormatter('%(message)s', level_formatter = self.fmt_level, time_formatter = self.fmt_time)
        # Done
        self.__change_handlers_formatter(fmt)


""" Setup a header to print when nodus starts """
def get_header():
    sp = 15
    s = []
    prob = 0.5
    for _ in range(3):
        ss = " "
        for i in range(sp):
            # Generate random number betwen 2190 and 21ff in hex 
            #
            # check if random prob is less than 0.3
            if np.random.rand() <= prob:
                #j = np.random.randint(0x2190, 0x25ff)
                #j = np.random.randint(0x25a0, 0x25ff)
                #j = np.random.randint(0x239b, 0x23a0)
                #j = np.random.randint(0x2500, 0x257f)
                j = np.random.randint(0x2596, 0x259f)
                #j = 0x232c
                ss += chr(j)
                #ss += "∘"
            else:
                ss += " "

        ss += " "
        s.append(ss)
    HEADER  = f"\x1B[48;2;141;47;102m\x1B[48;2;141;47;102m \x1B[48;2;117;39;85m \x1B[48;2;94;31;68m \x1B[48;2;70;23;51m \x1B[48;2;46;15;33m \x1B[48;2;23;7;17m \x1B[48;2;0;0;0m \x1B[48;2;23;7;17m \x1B[48;2;46;15;33m \x1B[48;2;70;23;51m \x1B[48;2;94;31;68m \x1B[48;2;117;39;85m \x1B[48;2;141;47;102m      \x1b[48;2;249;230;207m{s[0]}\x1B[0m\n"
    HEADER += f"\x1B[48;2;141;47;102m\x1B[48;2;141;47;102m \x1B[48;2;117;39;85m \x1B[48;2;94;31;68m \x1B[48;2;70;23;51m𐌍\x1B[48;2;46;15;33m \x1B[48;2;23;7;17m↠\x1B[48;2;0;0;0m⌾\x1B[48;2;23;7;17m↞\x1B[48;2;46;15;33m \x1B[48;2;70;23;51m𐌃\x1B[48;2;94;31;68m \x1B[48;2;117;39;85m𐌖\x1B[48;2;141;47;102m 𐌔    \x1B[48;2;250;198;122m{s[1]}\x1B[0m\n"
    HEADER += f"\x1B[48;2;141;47;102m\x1B[48;2;141;47;102m \x1B[48;2;117;39;85m \x1B[48;2;94;31;68m \x1B[48;2;70;23;51m \x1B[48;2;46;15;33m \x1B[48;2;23;7;17m \x1B[48;2;0;0;0m \x1B[48;2;23;7;17m \x1B[48;2;46;15;33m \x1B[48;2;70;23;51m \x1B[48;2;94;31;68m \x1B[48;2;117;39;85m \x1B[48;2;141;47;102m      \x1B[48;2;250;198;122m{s[2]}\x1B[0m\n"
    return HEADER

""" Get the next available name for a nodus database """
def get_next_name(base, existing_names):
    """Get the next available name for a nodus database."""
    # Get the next available name (nodus_db_1, nodus_db_2, ...)
    # even if an index doesn't exist (like if nodus_db_2 is there, but nodus_db_1 is not)
    # we should always pick the next available index (in this case 3, even though 1 isn't there)
    if base is None:
        base = 'nodus_db'
    if existing_names is None:
        existing_names = []

    # Create a regex pattern to match the base name
    pat = base.replace('_','\_')
    pat += r"\_?(\d+)?"
    pat = re.compile(pat)

    # Now let's find the highest number in the entries
    max_num = 0
    for entry in existing_names:
        match = pat.match(entry)
        if match:
            num = match.group(1)
            if num is not None:
                num = int(num)
                max_num = max(max_num, num)

    if max_num == 0:
        return base
    elif max_num > 0:
        return f"{base}_{max_num + 1}"

# Check if a process is running 
def can_pid_be_killed(pid):
    """Check if a process with the given PID is running."""
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return

def is_pid_running_proc(pid):
    """Check if a process with the given PID is running using /proc."""
    return os.path.exists(f"/proc/{pid}")

def is_pid_running_ps(pid):
    """Check if a process with the given PID is running using 'ps'."""
    try:
        result = subprocess.run(['ps', '-p', str(pid)], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return result.returncode == 0  # Return 0 means the process exists
    except Exception as e:
        print(f"[ERROR] Failed to check PID using 'ps': {e}")
        return False

# Define function is_process_running according to OS 
def check_platform():
    system = platform.system()

    if system == "Darwin":
        return "macOS"
    elif system == "Windows":
        return "Windows"
    elif system == "Linux":
        return "Linux"
    else:
        return "Unknown"

# Current system
current_system = check_platform()

is_pid_running = {'Linux': is_pid_running_proc, 'macOS': is_pid_running_ps, 'Windows': is_pid_running_ps}.get(current_system, is_pid_running_ps)


def kill_process(pid):
    # Check if the process is running
    if is_pid_running(pid):
        nodus.__logger__.error(f"Attempted to kill PID {pid}, but it is not running.")

    # Kill the process
    try:
        os.kill(pid, 9)
        nodus.__logger__.info(f"Job running at PID {pid} killed.")
        return True
    except Exception as e:
        nodus.__logger__.error(f"Failed to kill job at PID {pid}: {e}")
        return False


def compute_runtime(start_time, end_time):
    start_time = datetime.strptime(start_time, '%Y-%m-%d %H:%M:%S')
    end_time = datetime.strptime(end_time, '%Y-%m-%d %H:%M:%S')
    runtime = end_time - start_time

    # Now format the runtime into <number_of_days> days, HH:MM:SS
    days = runtime.days
    hours, remainder = divmod(runtime.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    if days == 0:
        return f"{hours:02}:{minutes:02}:{seconds:02}"
    else:
        return f"{days} days, {hours:02}:{minutes:02}:{seconds:02}"


# Import nodus 
import sys
sys.path.append('../nodus')

# Import args 
import argparse

# Import time
import time

# Import JobManager
import nodus
from nodus import JobManager


def run_script(script_path, shell = 'bash'):

    # Create Nodus Session
    session = nodus.NodusSession()
    db = session.add_nodus_db()  # Add a default NodusDB instance

    # Get the JobManager from the db instance
    jm = db.job_manager

    # Create a new job to run the script
    job_id = jm.create_job(
        name="example_job",
        parent_caller="example_script_runner",
        job_type="script",
        script_path = script_path,
        shell = shell
    )

    # Let the JobManager monitor jobs (in practice, your UI would run alongside this)
    while True:
        time.sleep(1)  # Keep the main thread alive


def run_command(command):
    # Create Nodus Session
    session = nodus.NodusSession()
    db = session.add_nodus_db()  # Add a default NodusDB instance

    # Get the JobManager from the db instance
    jm = db.job_manager

    # Create a new job to run the command
    job_id = jm.create_job(
        name="example_job",
        parent_caller="example_command_runner",
        job_type="command",
        command = command
    )

    # Let the JobManager monitor jobs (in practice, your UI would run alongside this)
    while True:
        time.sleep(1)  # Keep the main thread alive


def test_job_creation():
    # Initialize JobManager
    job_manager = JobManager()

    # Create a sample job
    job_id = job_manager.create_job(parent_caller='test', 
                                    status='pending', 
                                    log_path='/logs/test.log', 
                                    pid=None, 
                                    config='default_config')

    # Assert that the job ID is not None
    assert job_id is not None, "Job ID should not be None"

    # Wait for some time
    time.sleep(5)

    # Update job status to 'completed'
    job_manager.update_job_status(job_id, status='completed', completion_time=None)

    # Verify the job status in the database
    db_path = nodus.__nodus_db_path__
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute("SELECT status FROM jobs WHERE job_id = ?", (job_id,))
    status = cursor.fetchone()[0]
    conn.close()

    assert status == 'completed', f"Expected job status to be 'completed', but got '{status}'"

    # Cleanup: Delete the test job
    job_manager.delete_job(job_id)


def test_run_command(command, shell='bash'):
    # Initialize JobManager
    job_manager = JobManager()

    # Create a sample job
    job_id = job_manager.create_job(parent_caller='test', 
                                    status='pending', 
                                    log_path='/logs/test.log', 
                                    pid=None, 
                                    config='default_config')

    # Assert that the job ID is not None
    assert job_id is not None, "Job ID should not be None"






if __name__ == "__main__":

    # Parse arguments
    parser = argparse.ArgumentParser(description='Test Nodus Job Manager')

    # Add arguments
    parser.add_argument('-t', '--test', type=str, help='Test to run')
    parser.add_argument('--script', type=str, help='Script to run')
    parser.add_argument('--command', type=str, help='Command to run')
    parser.add_argument('--shell', type=str, help='Shell to use (e.g., bash, zsh)')

    # Parse arguments
    args = parser.parse_args()

    # Get shell
    shell = args.shell
    test = args.test
    script_path = args.script
    command = args.command

    # Run the test
    if test == 'job_creation':
        test_job_creation()
    elif test == 'script':
        script_path = 'tests/example_script.sh' if script_path is None else script_path
        run_script(script_path, shell=shell)
    elif test == 'command':
        nodus_command = "bash -c 'for i in {1..10}; do echo Step $i; sleep 1; done'"
        command = nodus_command if command is None else command
        run_command(command)


# nodus - Job Management Framework
# Author: Manuel Blanco Valentin
# Email: manuel.blanco.valentin@gmail.com
# Created: 2024-12-25
#
# Description:
# This file contains the Nodus class for managing the database of jobs.
# It provides methods for creating, inserting, updating, and querying job information.
#
# File: database.py
# Purpose:
# - Contains the Nodus class for interacting with the SQLite database
#   used to store and track job data, including job configurations,
#   statuses, and metadata.

""" Basic imports """
import os

# UUID 
import uuid

# Time management
from datetime import datetime

""" Database imports """
import sqlite3

# Import nodus itself
import nodus
from .manager import JobManager
from . import utils

class NodusSession:
    def __init__(self):
        # Generate a unique session ID
        self.session_id = str(uuid.uuid4())

        """Initialize a new Nodus session."""
        self.start_time = datetime.now()
        self.log_file = nodus.__nodus_log_file__
        self.nodus_dbs = {}
        self._keys = []

    # Get item 
    def __getitem__(self, key):
        if isinstance(key, str):
            if key in self.nodus_dbs:
                return self.nodus_dbs[key]
            else:
                # Log 
                nodus.__logger__.error(f"No NodusDB instance with name '{key}' found.")
                raise KeyError(f"No NodusDB instance with name '{key}' found.")
        elif isinstance(key, int):
            if key < len(self._keys):
                return self.nodus_dbs[self._keys[key]]
            else:
                # Log
                nodus.__logger__.error(f"Index {key} out of range.")
                raise IndexError(f"Index {key} out of range.")

    def add_nodus_db(self, name = None, db_path=None, create_table=True):
        # If name is none, get next available name (nodus_db_1, nodus_db_2, ...)
        if name is None:
            name = utils.get_next_name('nodus_db', self._keys)
        """Add a new NodusDB instance to the session."""
        nodus_db = NodusDB(name, self.session_id, db_path = db_path, create_table = create_table) # This also initializes the db table
        # Add to the list of NodusDB instances
        self.nodus_dbs[name] = nodus_db
        self._keys.append(name)

        # Log 
        nodus.__logger__.info(f"Added NodusDB instance '{name}' linked to database '{nodus_db.db_path}'")

        return nodus_db

    def close(self):
        """Close all NodusDB connections in the session."""
        for nodus_db in self.nodus_dbs.values():
            nodus_db.close()
        
    # Representation 
    def __repr__(self):
        s = f"NodusSession (session_id: {self.session_id})\n"
        s += f"   - Start time: {self.start_time}\n"
        s += f"   - Log file: {self.log_file}\n"
        s += f"   - NodusDB instances:\n"
        if len(self.nodus_dbs) > 0:
            for name, nodus_db in self.nodus_dbs.items():
                s += f"       - {name}: {nodus_db.db_path}\n"
        else:
            s += "       - None\n"
        
        return s


class NodusDB:
    def __init__(self, name, session_id, db_path=None, create_table = True):
        """Initialize the NodusDB database."""
        self.name = name
        # Set the parent session ID
        self.session_id = session_id

        self.db_path = db_path or nodus.__nodus_db_path__
        self.conn = sqlite3.connect(self.db_path)
        self.cursor = self.conn.cursor()

        # Create the jobs table if it doesn't exist
        if create_table: self.create_jobs_table()

        # Initialize the JobManager
        jm_name = f"{name}_job_manager"
        self.job_manager = JobManager(jm_name, self.db_path, self.session_id)

    def create_jobs_table(self):
        """Create the jobs table if it doesn't exist."""
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS jobs (
                job_id INTEGER PRIMARY KEY AUTOINCREMENT,
                nodus_session_id TEXT NOT NULL,
                parent_caller TEXT NOT NULL,
                job_name TEXT,
                status TEXT NOT NULL,
                timestamp TEXT NOT NULL,
                completion_time TEXT,
                log_path TEXT,
                pid TEXT DEFAULT NULL,
                config TEXT
            )
        ''')
        self.conn.commit()
        # Log 
        nodus.__logger__.info(f"Created jobs table in NodusDB instance '{self.name}'")

    def close(self):
        """Close the database connection."""
        self.conn.close()
        # Log 
        nodus.__logger__.info(f"Closed NodusDB instance '{self.name}'")
    
    def __repr__(self):
        s = f'<NodusDB {self.name} - {self.db_path}>'
        # Add JobManager info
        s += f'\n    - JobManager: {self.job_manager}'
        return s


# Example usage
if __name__ == "__main__":
    
    import random

    # Initialize NodusSession
    session = NodusSession()
    print(session)
    db = session.add_nodus_db()  # Add a default NodusDB instance
    print(db)

    # Get the JobManager from the db instance
    jm = db.job_manager

    # Get a random letter (A,B,C,..)
    parent_caller = 'project_' + random.choice('ABCDEFGHIJKLMNOPQRSTUVWXYZ')
    # Random PID number with 5 digits
    pid = random.randint(10000, 99999)

    # Insert a new job with some config
    config = {"param1": 42, "param2": "value"}
    job_type = 'command'
    job_id, job = jm.create_job(parent_caller, job_type, command=None, script_path=None, pid=None)

    # Query jobs
    jobs = jm.get_jobs(parent_caller=parent_caller)
    print(jobs)

    # Retrieve job config
    #job_config = nodus.get_job_config(1)
    #print(job_config)

    # Close connection
    session.close()


# Basic imports 
import os 

""" First of all, let's setup the directory where nodus is located """
__nodus_bin_dir__ = os.path.dirname(os.path.abspath(__file__))
__version__ = "0.1.0"

""" Also get the user's home directory where we will store data and also logs """
__nodus_data_dir__ = os.path.expanduser('~/.nodus')
# Make sure dir exists 
try:
    os.makedirs(__nodus_data_dir__, exist_ok=True)
except Exception as e:
    logging.error(f"Failed to create data directory: {e}")

# Setup db path
__nodus_db_path__ = os.path.join(__nodus_data_dir__, 'nodus.db')

""" Import utils """
from .utils import CustomLogger, get_header
# Print header
HEADER = get_header()
print(HEADER)

""" Setup logger """
__logger__ = CustomLogger(dir = __nodus_data_dir__)
__logger__.info("Nodus initialized")
__logger__.info(f"Nodus version: {__version__}")

# Get log file path
__nodus_log_file__ = __logger__.log_file_path

# We can initialize the log dir (where we will put the logs for each job, in case the user doesn't pass one)
__nodus_logs_dir__ = os.path.join(__nodus_data_dir__, 'logs')
if not os.path.exists(__nodus_logs_dir__):
    os.makedirs(__nodus_logs_dir__)
    __logger__.info(f"Created logs directory: {__nodus_logs_dir__}")

# Import nodus
from .db import NodusSession, NodusDB
__logger__.info("Nodus imported")
# Import JobManager
from .manager import JobManager, Job
__logger__.info("JobManager imported")
__logger__.info("Nodus ready to use")

# Import ui if needed
try:
    from .ui import curses_main
except ImportError as e:
    __logger__.error(f"Failed to import UI module: {e}")

""" At exit to determine when the program is closing """
import atexit

""" Register the final logging message when exiting the program """
def log_before_exit():
    __logger__.close()

# Register the function to be called on exit
atexit.register(log_before_exit)



# nodus - Job Management Framework
# Author: Manuel Blanco Valentin
# Email: manuel.blanco.valentin@gmail.com
# Created: 2024-12-25
#
# Description:
# This file contains the JobManager class responsible for managing the 
# execution of jobs, including tracking their status, job queue, and 
# interacting with the database.
#
# File: manager.py
# Purpose:
# - Contains the JobManager class for handling job queues, statuses,
#   and interacting with the Nodus database to track job progress.

# Basic modules
import os

# Time control 
import time
from datetime import datetime

# SQLite database
import sqlite3

# Import nodus
import nodus

""" Threading """
import subprocess
import threading

""" tempfile for temporary files creation """
import tempfile

""" Generic Job class """
class Job:
    def __init__(self, name, job_id, nodus_session_id, parent_caller = "nodus", pid = None, config = None, status = "pending"):
        self.name = name
        self.job_id = job_id
        self.nodus_session_id = nodus_session_id

        # Parent caller
        self.parent_caller = parent_caller if parent_caller else "nodus"

        # PID, status and config
        self.pid = pid
        self.status = status
        self.config = config

        # Init process to None (only valid after running the subprocess)
        self.process = None

        # Init markers to track progress of processes 
        self.marker_dir = tempfile.gettempdir()  # Use system's temp directory
        self.start_marker = os.path.join(self.marker_dir, f"nodus_{nodus_session_id}_{job_id}_start")
        self.end_marker = os.path.join(self.marker_dir, f"nodus_{nodus_session_id}_{job_id}_end")
        self.has_started = False  # Tracks if the start marker has been detected
        self.has_ended = False  # Tracks if the end marker has been detected

    # Even though this is defined here, it's only called in the children classes because we set the job_type there
    def _create_log_path(self):
        """Create a log file path for the job."""
        log_dir = os.path.join(nodus.__nodus_logs_dir__, self.parent_caller, self.nodus_session_id)
        os.makedirs(log_dir, exist_ok=True)
        log_name = f'job_{self.job_type}_{self.job_id}_{self.name}'
        log_name += '.log' if not self.pid else f'_pid{self.pid}.log'
        log_file = os.path.join(log_dir, log_name)
        return log_file
    
    """ This function helps us keep track of processes by creating a temporary marker file. 
        One file will be created at the start of the process, and another one at the end. 
        This, along with checking for whether the PID is running or not, allows us 
        to determine the status of the process accurately.
    """
    def _create_marker_file(self, marker_path):
        """Create a marker file to indicate job state."""
        if not os.path.exists(marker_path):
            with open(marker_path, 'w') as f:
                f.write("")
        # Log 
        nodus.__logger__.info(f"Marker file created for job {self.job_id}: {marker_path}")
        return marker_path

    def _check_job_status(self):
        """Determine the current status of a job."""
        # If the job was started as a process, check its return code
        if self.process:
            return_code = self.process.poll()  # Non-blocking check for completion
            if return_code is None:
                self.status = "running"
            else:
                self.status = "completed" if return_code == 0 else "errored"
                self.finalize()  # Create end marker
            return self.status

        # If no process, fall back to markers and PID checks
        if not self.has_started:
            if os.path.exists(self.start_marker):
                self.status = "running"
                self.has_started = True
                return self.status

        if not os.path.exists(self.start_marker):
            if nodus.utils.is_pid_running(self.pid):
                self.status = "running"
            elif os.path.exists(self.end_marker):
                self.status = "completed"
            else:
                self.status = "errored"
        return self.status
    
    def finalize(self):
        """Create end marker file."""
        self._create_marker_file(self.end_marker)

    """ Representation """
    def __repr__(self):
        s = f'<Job {self.job_id} - {self.job_type}>'
        s += f'\n    - Status: {self.status}'
        s += f'\n    - Log: {self.log_path}'
        if self.config is not None:
            if len(self.config) > 0:
                s += f'\n    - Config: '
                for key, value in self.config.items():
                    s += f'\n        - {key}: {value}'
        if self.pid:
            s += f'\n    - PID: {self.pid}'
        return s 
        

""" Command Job class """
class CommandJob(Job):
    def __init__(self, name, job_id, nodus_session_id, command = None, log_path = None, **kwargs):
        # Call super to set the basic attributes
        super().__init__(name, job_id, nodus_session_id, **kwargs)

        # Set the job type
        self.job_type = "command"

        # Set the command & shell
        self.command = command

        # Init the path
        self.log_path = log_path if log_path else self._create_log_path()
    
    def run(self):
        # Create start marker file
        start_marker = self._create_marker_file(self.start_marker)

        if self.command:
            with open(self.log_path, 'a') as log_file:
                self.process = subprocess.Popen(
                    self.command, shell=True, stdout = log_file, stderr = subprocess.STDOUT
                    )
            self.pid = self.process.pid
            # Log 
            nodus.__logger__.info(f"Command Job {self.job_id} started with PID: {self.pid}")
        else:
            nodus.__logger__.error(f"Command not provided for Job {self.job_id}.")
    
    """ Repr """
    def __repr__(self):
        s = super().__repr__()
        if self.command:
            s += f'\n    - Command: {self.command}'
        return s


""" Script Job class """
class ScriptJob(Job):
    def __init__(self, name, job_id, nodus_session_id, script_path = None, log_path = None, shell = 'bash', **kwargs):
        # Call super to set the basic attributes
        super().__init__(name, job_id, nodus_session_id, **kwargs)

        # Set the job type
        self.job_type = "script"

        # Set the script path
        self.script_path = script_path
        self.shell, self.shell_flag = self._process_shell(shell)

        # Init the path
        self.log_path = log_path if log_path else self._create_log_path()
    
    def _process_shell(self, shell):
        if shell is None:
            return 'bash', '-c'
        
        shell_flags = {
            'fish': '-e',
            'bash': '-c',
            'zsh': '-c',
            'sh': '-c',
            'ksh': '-c',
            'tcsh': '-c',
            'csh': '-c',
            'dash': '-c',  # Add support for Dash shell,
            'python': '-m'
        }
        return shell if shell in shell_flags else 'bash', shell_flags.get(shell, 'c')

    def run(self):
        # Create start marker file
        start_marker = self._create_marker_file(self.start_marker)

        if self.script_path:
            with open(self.log_path, 'a') as log_file:
                self.process = subprocess.Popen(
                    [self.shell, self.script_path], stdout = log_file, stderr = subprocess.STDOUT
                    )
            self.pid = self.process.pid
            # Log 
            nodus.__logger__.info(f"Script Job {self.job_id} started with PID: {self.pid}")
        else:
            nodus.__logger__.error(f"Script path not provided for Job {self.job_id}.")
    
    """ Repr """
    def __repr__(self):
        s = super().__repr__()
        if self.script_path:
            s += f'\n    - Script: {self.script_path}'
        if self.shell:
            s += f'\n    - Shell: {self.shell}'
        return s

""" Class for Jobs that adopt an existing PID """
class AdoptedPIDJob(Job):
    def __init__(self, name, job_id, nodus_session_id, pid = None, log_path = None, **kwargs):
        # Call super to set the basic attributes
        super().__init__(name, job_id, nodus_session_id, pid = pid, **kwargs)

        # Set the job type
        self.job_type = "pid"

        # Set the PID
        self.pid = pid

        # Init the path
        self.log_path = self._create_log_path() if log_path is None else log_path

    def run(self):
        # Create start marker file
        start_marker = self._create_marker_file(self.start_marker)

        if self.pid:
            nodus.__logger__.info(f"Adopting process with PID: {self.pid}")
            return  # No need to spawn a new process in this case.

        if self.process:
            self.pid = self.process.pid

    """ Repr """
    def __repr__(self):
        s = super().__repr__()
        return s


""" Job Manager class """
class JobManager:
    def __init__(self, name: str, db_path: str, nodus_session_id: str):
        """Initialize the JobManager with a NodusDB instance."""
        self.name = name
        self.db_path = db_path
        self.nodus_session_id = nodus_session_id

        # Create a sql connection
        self.conn = sqlite3.connect(self.db_path)
        
        # Keep track of running PIDs and job_ids
        self.running_pids = {}  
        self.jobs = {}  # Store all Job objects by job_id
        self._keys = [] # Store all job names

        # Start the monitor thread which will check the status of all jobs
        self.monitor_thread = threading.Thread(target=self._monitor_all_jobs, daemon=True)
        self.monitor_thread.start()

    def _create_job_entry(self, name, parent_caller, job_type, nodus_session_id, command = None, script_path = None, status='pending', log_path=None, pid=None, config=None, **kwargs):
        
        """Create a new job in the database and handle job execution."""
        timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        # Make sure pid is text
        pid = f"{pid}" if pid else None

        # Create job entry in the database
        # Create a cursor 
        cursor = self.conn.cursor()
        cursor.execute('''
            INSERT INTO jobs (nodus_session_id, parent_caller, job_name, status, timestamp, log_path, pid, config)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?)
        ''', (nodus_session_id, parent_caller, name, status, timestamp, log_path, pid, config))
        self.conn.commit()

        job_id = cursor.lastrowid

        # Log 
        nodus.__logger__.info(f"Job {job_id} created.")

        return job_id
    
    def create_job(self, parent_caller, job_type, name = None, nodus_session_id = None, **kwargs):
        """Create and run a new job."""
        # Check name 
        if name is None:
            name = nodus.utils.get_next_name(self.name.replace('_job_manager','_job'), self._keys)
        if nodus_session_id is None:
            nodus_session_id = self.nodus_session_id

        # Create job entry in the database
        job_id = self._create_job_entry(name, parent_caller, job_type, nodus_session_id, **kwargs)
        
        # Create the Job object
        job_class = {'command': CommandJob, 'script': ScriptJob, 'pid': AdoptedPIDJob}.get(job_type, Job)
        job = job_class(name, job_id, nodus_session_id, **kwargs)

        # Update log_path 
        if job.log_path is not None:
            kwargs['log_path'] = job.log_path
            # Update 
            self.update_log_path(job_id, job.log_path)
        
        # Run the job
        job.run()

        # Track the running PID and store the Job object
        if job.pid:
            self.running_pids[job.pid] = job_id
            # Also update pid in table 
            cursor = self.conn.cursor()
            cursor.execute('''
                UPDATE jobs
                SET pid = ?
                WHERE job_id = ?
            ''', (str(job.pid), job_id))
            self.conn.commit()
            # Log 
            nodus.__logger__.info(f"Job {job_id} started with PID: {job.pid}")

        self.jobs[job_id] = job
        self._keys.append(name)
        
        return job_id, job

    def update_job_status(self, job_id, status, job_pid, completion_time=None, db_conn = None):
        """Update the status of a job."""
        completion_time = completion_time or datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        if db_conn is None:
            nodus.__logger__.info("No db connection, picking from self")
            cursor = self.conn
        else:
            cursor = db_conn

        job_pid_indicator = f"{job_pid}*" if job_pid else None

        cursor = db_conn.cursor()
        cursor.execute('''
            UPDATE jobs
            SET status = ?, completion_time = ?, pid = ?
            WHERE job_id = ?
        ''', (status, completion_time, job_pid_indicator, job_id))
        db_conn.commit()

        # Log status updated 
        nodus.__logger__.info(f"Job {job_id} status updated to {status} and PID to {job_pid_indicator}.")

    def update_log_path(self, job_id, log_path):
        """Update the log path of a job."""
        cursor = self.conn.cursor()
        cursor.execute('''
            UPDATE jobs
            SET log_path = ?
            WHERE job_id = ?
        ''', (log_path, job_id))
        self.conn.commit()

        # Log 
        nodus.__logger__.info(f"Job {job_id} log path updated to {log_path}.")

    def delete_job(self, job_id):
        """Delete a job from the database."""
        cursor = self.conn.cursor()
        cursor.execute('''
            DELETE FROM jobs
            WHERE job_id = ?
        ''', (job_id,))
        self.conn.commit()

        # Log job deleted
        nodus.__logger__.info(f"Job {job_id} deleted.")
    
    def kill_job(self, job_id):
        
        """Kill a running job."""
        job = self.jobs.get(job_id)
        if job is None:
            nodus.__logger__.error(f"Job {job_id} not found.")
            return False

        if job.pid is None:
            nodus.__logger__.error(f"Job {job_id} has no PID.")
            return False

        # If pid has the * at the end, skip cause this is an old PID and the process already ended
        if job.pid.endswith("*"):
            nodus.__logger__.error(f"Job {job_id} with PID {job.pid} already ended.")
            return False

        # Check if the process is running
        if not nodus.utils.is_pid_running(job.pid):
            nodus.__logger__.error(f"Job {job_id} with PID {job.pid} is not running.")
            # Update table to add the * at the end of the PID 
            cursor = self.conn.cursor()
            cursor.execute('''
                UPDATE jobs
                SET pid = ?
                WHERE job_id = ?
            ''', (f"{job.pid}*", job_id))
            self.conn.commit()
            return False

        # Kill the process
        try:
            os.kill(job.pid, 9)
            nodus.__logger__.info(f"Job {job_id} killed.")
            return True
        except Exception as e:
            nodus.__logger__.error(f"Failed to kill job {job_id}: {e}")
            return False

    # get a list of jobs 
    def get_jobs(self):
        """Get a list of all jobs."""
        cursor = self.conn.cursor()
        cursor.execute('''
            SELECT job_id, nodus_session_id, parent_caller, job_name, status, timestamp, completion_time, log_path, pid, config
            FROM jobs
            ORDER BY job_id DESC
        ''')
        jobs = cursor.fetchall()

        return jobs

    def get_job(self, job_id):
        """Get a specific job by its ID."""
        cursor = self.conn.cursor()
        cursor.execute('''
            SELECT job_id, nodus_session_id, parent_caller, job_name, status, timestamp, completion_time, log_path, pid, config
            FROM jobs
            WHERE job_id = ?
        ''', (job_id,))
        job = cursor.fetchone()

        return job
    
    # Monitor all jobs 
    def _monitor_all_jobs(self):
        """ Note: this process is run in a separate thread,
            which means we cannot use the same db sqlite connection
            we created earlier. We need to create a new connection 
            and pass it to the update_job_status function
        """
        conn = sqlite3.connect(self.db_path)

        """Monitor all running jobs."""
        while True:
            finished_pids = []
            for job_id, job in self.jobs.items():
                if job.status in ['completed', 'errored']:
                    continue # Skip already completed jobs
                
                # check the status of the job
                current_status = job._check_job_status()

                if current_status in ['completed', 'errored']:
                    # Update the job status to completed
                    self.update_job_status(job_id, current_status, job.pid, db_conn = conn)
                    # Remove the PID from the running list
                    finished_pids.append(job.pid)

            # Remove finished jobs from the running list
            for pid in finished_pids:
                if pid in self.running_pids:
                    del self.running_pids[pid]

            time.sleep(1)
    



import curses
import sqlite3
import os
from datetime import datetime
import time

import nodus

# Import subprocess to run 'less' command
import subprocess

def curses_main(stdscr):
    """Main curses function."""
    curses.curs_set(0)  # Hide the cursor
    stdscr.nodelay(True)  # Make getch() non-blocking
    stdscr.clear()

    # Colors
    curses.start_color()
    curses.init_pair(1, curses.COLOR_GREEN, curses.COLOR_BLACK)
    curses.init_pair(2, curses.COLOR_RED, curses.COLOR_BLACK)
    curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)
    curses.init_pair(4, curses.COLOR_CYAN, curses.COLOR_BLACK)

    # Initialize a nodus session
    session = nodus.NodusSession()
    db = session.add_nodus_db()  # Add a default NodusDB instance

    # Get the JobManager from the db instance
    jm = db.job_manager

    current_row = 0
    last_update_time = datetime.now()
    current_time = last_update_time.strftime('%H:%M:%S')
    header = f"Nodus Job Manager (Standalone Mode) - {current_time}"

    while True:
        # Periodically refresh the header with the current time
        now = datetime.now()
        if (now - last_update_time).seconds >= 1:
            last_update_time = now
            current_time = now.strftime('%H:%M:%S')
            header = f"Nodus Job Manager (Standalone Mode) - {current_time}"

        # Get jobs
        jobs = jm.get_jobs()

        # Clear and redraw the main screen
        stdscr.clear()
        height, width = stdscr.getmaxyx()

        # Header
        stdscr.attron(curses.color_pair(4))
        stdscr.addstr(0, width // 2 - len(header) // 2, header)
        stdscr.attroff(curses.color_pair(4))

        # Instructions
        instructions = "Use UP/DOWN to navigate, ENTER to select, D to delete, K to kill, L to open log, Q to quit"
        stdscr.addstr(1, 0, instructions, curses.color_pair(3))

        # Display jobs
        if jobs:
            stdscr.addstr(3, 0, f"{'ID':<5} {'PID':<8} {'NODUS_SESSION_ID':<40} {'PARENT CALLER':<30} {'JOB NAME':<15} {'STATUS':<15} {'START TIME':<20} {'END TIME':<20} {'RUNTIME':<15} {'LOG':<30}", curses.A_BOLD)
            for idx, (job_id, nodus_session_id, parent_caller, job_name, status, timestamp, completion_time, log_path, pid, config) in enumerate(jobs):
                # If completion_time is None, display "N/A"
                completion_time = completion_time if completion_time else "N/A"
                # If PID is None, display "N/A"
                pid = pid if pid else "N/A"
                # If parent_caller is None, display "N/A"
                parent_caller = parent_caller if parent_caller else "N/A"
                # If parent_caller is too long, truncate it
                parent_caller = parent_caller[:27] + "..." if len(parent_caller) > 30 else parent_caller
                # If job_name is None, display "N/A"
                job_name = job_name if job_name else "N/A"
                # If job_name is too long, truncate it
                job_name = job_name[:12] + "..." if len(job_name) > 15 else job_name
                # If endtime is not N/A, calculate runtime
                runtime = "N/A"
                if completion_time != "N/A" and timestamp:
                    runtime = nodus.utils.compute_runtime(timestamp, completion_time)
                # If log_path is None, N/A
                log_path = log_path if log_path else "N/A"
                # If log_path is too long, truncate it
                if len(log_path) > 20:
                    lp = log_path.split(os.sep)[-1]
                    log_path = f'…{os.sep}{lp[:15]}…{lp[-14:]}' if len(lp) > 30 else f"…{os.sep}{lp}"
                #log_path = log_path[:17] + "..." if len(log_path) > 20 else log_path
                # Highlight the current row
                if idx == current_row:
                    stdscr.attron(curses.color_pair(1))
                    stdscr.addstr(4 + idx, 0, f"{job_id:<5} {pid:<8} {nodus_session_id:<40} {parent_caller:<30} {job_name:<15} {status:<15} {timestamp:<20} {completion_time:<20} {runtime:<15} {log_path:<30}")
                    stdscr.attroff(curses.color_pair(1))
                else:
                    stdscr.addstr(4 + idx, 0, f"{job_id:<5} {pid:<8} {nodus_session_id:<40} {parent_caller:<30} {job_name:<15} {status:<15} {timestamp:<20} {completion_time:<20} {runtime:<15} {log_path:<30}")
        else:
            stdscr.addstr(3, 0, "No jobs found.", curses.color_pair(2))

        stdscr.refresh()

        # Handle key inputs
        try:
            key = stdscr.getch()
        except:
            key = None

        if key == curses.KEY_UP and current_row > 0:
            current_row -= 1
        elif key == curses.KEY_DOWN and current_row < len(jobs) - 1:
            current_row += 1
        elif key == ord("q"):
            break
        elif key == ord("d") and jobs:
            # Delete the selected job entry from the database
            selected_job = jobs[current_row]
            confirm = display_confirmation_popup(stdscr, "Delete this job entry from database? This will NOT kill the process. (y/n)")
            if confirm:
                jm.delete_job(selected_job[0])
        elif key == ord("k") and jobs:
            selected_job = jobs[current_row]
            # Get PID 
            selected_pid = selected_job[8]
            # if PID is N/A or contains an asterisk, tell the user this process cannot be killed 
            skip = False
            if selected_pid == 'N/A':
                skip = True
                display_message_notification(stdscr, "This process cannot be killed. PID is NULL.")
            if selected_pid.endswith('*'):
                display_message_notification(stdscr, "This process cannot be killed since the process already finished (note the * next to the PID number).")
                skip = True
            
            if not skip:
                confirm = display_confirmation_popup(stdscr, f"Kill this job with PID {selected_pid}? This will NOT delete entry from database. (y/n)")
                if confirm:
                    process_killed = nodus.utils.kill_process(selected_pid)
                    # If process_killed is False, show an error message to user
                    # Make sure to delete the previous confirmation popup 
                    stdscr.clear()
                    if not process_killed: display_message_notification(stdscr, "Failed to kill the process!")

        elif key == ord("l") and jobs:
            selected_job = jobs[current_row]
            log_path = selected_job[7]
            if log_path is None:
                display_message_notification(stdscr, f"Log file is null.")
            else:
                if os.path.exists(log_path):
                    # Check if process is still running, if it is, open with tail -F, else open with less
                    status = selected_job[4]
                    if status in ['completed','error','killed']:
                        # Just open file in less
                        open_file_in_less(stdscr, log_path)
                    else:
                        open_file_in_tail(stdscr, log_path)
                else:
                    stdscr.clear()
                    display_message_notification(stdscr, f"Log file {log_path} not found!")
        elif key == ord("\n") and jobs:
            selected_job = jobs[current_row]
            handle_job_selection(stdscr, selected_job)

        time.sleep(0.1)  # Small delay to avoid high CPU usage

def display_confirmation_popup(stdscr, message):
    """Display a confirmation popup."""
    height, width = stdscr.getmaxyx()
    popup_height = 3
    popup_width = len(message) + 4
    popup_y = height // 2 - popup_height // 2
    popup_x = width // 2 - popup_width // 2

    # Draw the popup
    stdscr.attron(curses.color_pair(3))
    stdscr.addstr(popup_y, popup_x, f"+{'-' * (popup_width - 2)}+")
    stdscr.addstr(popup_y + 1, popup_x, f"| {message} |")
    stdscr.addstr(popup_y + 2, popup_x, f"+{'-' * (popup_width - 2)}+")
    stdscr.attroff(curses.color_pair(3))
    stdscr.refresh()

    while True:
        key = stdscr.getch()
        if key == ord("y"):
            return True
        elif key == ord("n"):
            return False
        
def display_message_notification(stdscr, message):
    """Display a message notification."""
    height, width = stdscr.getmaxyx()
    popup_height = 3
    popup_width = len(message) + 4
    popup_y = height // 2 - popup_height // 2
    popup_x = width // 2 - popup_width // 2

    # Draw the popup
    stdscr.attron(curses.color_pair(3))
    stdscr.addstr(popup_y, popup_x, f"+{'-' * (popup_width - 2)}+")
    stdscr.addstr(popup_y + 1, popup_x, f"| {message} |")
    stdscr.addstr(popup_y + 2, popup_x, f"+{'-' * (popup_width - 2)}+")
    stdscr.attroff(curses.color_pair(3))
    stdscr.refresh()
    
    while True:
        # Wait until ENTER
        key = stdscr.getch()
        if key == ord("\n"):
            break

def handle_job_selection(stdscr, job):
    """Handle selection of a job."""
    stdscr.clear()
    height, width = stdscr.getmaxyx()

    # Display job details
    title = f"Job Details (ID: {job[0]})"
    stdscr.attron(curses.color_pair(4))
    stdscr.addstr(0, width // 2 - len(title) // 2, title)
    stdscr.attroff(curses.color_pair(4))

    details = [
        f"Nodus session ID: {job[1]}",
        f"Parent Caller: {job[2]}",
        f"Job Name: {job[3]}",
        f"Status: {job[4]}",
        f"Start Time: {job[5]}",
        f"End Time: {job[6] if job[6] else 'N/A'}",
        f"Log Path: {job[7]}",
        f"PID: {job[8]}",
        f"Config: {job[9]}",
    ]
    

    for idx, line in enumerate(details):
        stdscr.addstr(2 + idx, 0, line)

    instructions = "Press B or ENTER to go back | D to delete job | K to kill job | L to open log"
    stdscr.addstr(height - 2, 0, instructions, curses.color_pair(3))
    stdscr.refresh()

    while True:
        key = stdscr.getch()
        if key == ord("b") or key == ord("\n"):
            break
        elif key == ord("k"):
            confirm = display_confirmation_popup(stdscr, "Kill this job? (y/n)")
            if confirm:
                delete_job(job[0])
                break
        elif key == ord("l"):
            log_path = job[7]
            if log_path is not None:
                if os.path.exists(log_path):
                    open_file_in_less(stdscr, log_path)
                else:
                    stdscr.addstr(height - 3, 0, "Log file not found!", curses.color_pair(2))
                    stdscr.refresh()
                    time.sleep(1)
            else:
                stdscr.addstr(height - 3, 0, "Log file is null!", curses.color_pair(2))
                stdscr.refresh()
                time.sleep(1)


def open_file_in_less(stdscr, file):
    """Open the file in 'less' and return to the curses UI cleanly."""
    curses.endwin()  # Restore the terminal to its normal state
    try:
        subprocess.run(['less', file])  # Run 'less' on the file
    finally:
        curses.doupdate()  # Reinitialize the curses screen

def open_file_in_tail(stdscr, file):
    """Open the file in 'tail -F' and return to the curses UI cleanly.
        We need a new subprocess cause issuing a Control+C to kill tail
        will also kill the curses UI.
    """
    curses.endwin()  # Restore the terminal to its normal state
    try:
        subprocess.Popen(
            ["tail", "-F", file],
            preexec_fn=os.setpgrp  # Start a new process group
        ).wait()
    except KeyboardInterrupt:
        pass
    finally:
        curses.doupdate()  # Reinitialize the curses screen

    
def run_ui():
    curses.wrapper(curses_main)



from setuptools import setup, find_packages

setup(
    name="nodus",
    version="0.0.6",
    author="Manuel Blanco Valentin",
    author_email="manuel.blanco.valentin@gmail.com",
    description="Nodus - A lightweight and reusable job manager.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/manuelblancovalentin/nodus",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)